from cherrypy import _cphttptools
from cherrypy.lib import cptools
import types
import cherrypy
import cgi
import sys
import os

from subway import helpers

class NO_DEFAULT:
    pass

class Wrapper(object):
    """
    Wraps up dictionaries, internal class. Lets you use getattr
    just like getitem, and optionally return None instead of
    raising a KeyError.
    """
    def __init__(self, wrapped=None, default_return = NO_DEFAULT):
        if wrapped:
            self._d = wrapped
        else:
            self._d = {}
        if default_return is not NO_DEFAULT:
            self._default_return = default_return
        if not hasattr(self, "_default_return"):
            self._default_return = NO_DEFAULT
    def __delitem__(self, key):
        del self._d[key]
    def __getattribute__(self, key):
        try:
            return object.__getattribute__(self, key)
        except AttributeError:
            raise
        try:
            return self[key]
        except KeyError, e:
            raise AttributeError, e
    def __getitem__(self, key):
        try:
            return self._d[key]
        except KeyError:
            if self._default_return is not NO_DEFAULT:
                return self._default_return
            else:
                raise
    def __setitem__(self, key, value):
        self._d[key] = value
    def __setattr__(self, key, value):
        if key[0] == "_":
            object.__setattr__(self, key, value)
        else:
            self._d[key] = value
    def copy(self):
        return self._d.copy()
    def clear(self):
        return self._d.clear()
    def update(self, d):
        return self._d.update(d)
    def get(self, *args, **kwargs):
        return self._d.get(*args, **kwargs)
    def keys(self):
        return self._d.keys()
    def items(self):
        return self._d.items()
    def values(self):
        return self._d.values()
    def has_key(self, k):
        return self._d.has_key(k)
    def __contains__(self, k):
        return k in self._d
    def __repr__(self):
        return self._d.__repr__()
    def __str__(self):
        return self._d.__str__()
    def __len__(self):
        return len(self._d)

class Session(Wrapper):
    """
    A more user-friendly session object that allows the
    end user to use session.keyname instead of session["keyname"],
    though the latter still works.
    """
    _d = property(lambda self: cherrypy.session, lambda self, v: None)
    def __init__(self):
        Wrapper.__init__(self)

session = Session()

class Request(object):
    """
    The Subway request object.
    
    Properties:
    args - Wrapper dictionary of the request arguments (cherrypy.request.paramMap)
    headers - Wrapper dictionary of the incoming headers (cherrypy.request.headerMap)
    cookies - Wrapper dictionary of the incoming HTTP cookies (cherrypy.request.simpleCookie)
    path - the request URL path
    url - the URL the browser used to fetch this resource
    method - the HTTP method used to fetch this resource
    """
    args = property(lambda self: Wrapper(cherrypy.request.paramMap))
    headers = property(lambda self: Wrapper(cherrypy.request.headerMap))
    cookies = property(lambda self: cherrypy.request.simpleCookie)
    path = property(lambda self: cherrypy.request.objectPath or cherrypy.request.path)
    url = property(lambda self: cherrypy.request.browserUrl)
    base = property(lambda self: cherrypy.request.base)
    method = property(lambda self: cherrypy.request.method)
    remote_host = property(lambda self: cherrypy.request.remoteHost)
    remote_addr = property(lambda self: cherrypy.request.remoteAddr)
    remote_port = property(lambda self: cherrypy.request.remotePort)

request = Request()    


class Response(object):
    """
    The Subway response object.

    Properties:
    headers - Wrapper dictionary of the outgoing headers (cherrypy.response.headerMap)
    cookies - Wrapper dictionary of the outgoing cookies (cherrypy.response.simpleCookie)

    Methods:
    write(content) - write out content to the browser
    redirect(url) - redirect the user to a given URL. you should return this return value
                    from your function.
    serve_file(filename) - return this from your method to serve a static file
    """
    headers = property(lambda self: Wrapper(cherrypy.response.headerMap))
    cookies = property(lambda self: cherrypy.response.simpleCookie)
    
    def write(self, d):
        cherrypy.response.body.append(d)
    
    def escape(self, msg):
        return helpers.escape(msg)    
    
    def redirect(self, url):
        raise cherrypy.HTTPRedirect(url)
        
    def no_cache(self):
        cherrypy.response.headerMap["Expires"] = "Sat, 1 Jan 1970 12:00:00 GMT"
        cherrypy.response.headerMap["Cache-Control"] = "no-store, no-cache, must-revalidate, post-check=0, pre-check=0"
        cherrypy.response.headerMap["Pragma"] = "no-cache"

    def serve_file(self, filename):
        return cptools.serveFile(filename)

response = Response()        


# ---------------------------------------------------------------------

_FLASH_PREFIX = '_flash_'

#@@ inherit from Wrapper?  
#  one step at time, please!
class Flash(object):
    """ 
    A port of Ruby on Rails' Flash.
    (see ticket #91, http://subway.python-hosting.com/ticket/91)
    
    Flashes are simple dict-like objects primarily intended to be used in your
    Subway application to store status messages. 
    
    Best explained in the very Rails Action Pack docs [1]: 
        ''The flash provides a way to pass temporary objects between actions.
        Anything you place in the flash will be exposed to the very next action
        and then cleared out. This is a great way of doing notices and alerts,
        such as a create action that sets flash["notice"] = "Successfully
        created" before redirecting to a display action that can then expose
        the flash to its template.''
    [1] http://api.rubyonrails.com/classes/ActionController/Flash.html
    
    How to use it
    -------------
    
    Say you have an hanlder which saves an entry to your database. A common
    idiom is to redirect to another handler after the creation has occurred.
    Assuming that 'flash' is a global name referencing a Flash object, in the
    controller class we have do the following:
        
        @expose()
        def save_thing(thing_name):
            
            # ... saving code ...
            
            flash['notice'] = "Your thing %s has been saved" % thing_name
            return redirect("/list_things")
        
        @expose()
        def list_things():
            things = model.Thing.select()
            return view.list_things()
        
    Then, in your template:
    
        $widget.notice
        
        #for $thing in $things
            ## display the thing...
        #end for
         
    Now, if 'flash' were a normal dictionary, you'd have to be careful
    to delete the 'notice' item after it's used. Not with a Flash object.
    After the `list_things` action, the 'notice' key will be automatically
    deleted.
    """ 
                    
    def __setitem__(self, key, value):            
        key = _FLASH_PREFIX + key        
        response_cookie = cherrypy.response.simpleCookie
        
        response_cookie[key] = value
        response_cookie[key]['path'] = '/'   

    def __getitem__(self, key):
        """
        Retrieve the flash value, if one is set. 
        The value is cleared after retrieval.        
        """                       
        key = _FLASH_PREFIX + key
        
        request_cookie = cherrypy.request.simpleCookie
        response_cookie = cherrypy.response.simpleCookie
                
        try:
            value = request_cookie[key].value                                    
            # make to expire now            
            response_cookie[key] = ''
            response_cookie[key]['path'] = '/'   
            response_cookie[key]['expires'] = 0 #@@ change me to some past date
        except KeyError, exc:
            value = ''
        
        return value
    
    def __contains__(self, key):                
        request_cookie = cherrypy.request.simpleCookie
    
        key = _FLASH_PREFIX + key
        return key in request_cookie

    # "The current [Cheetah] NameMapper implementation does key lookup only 
    # if the object  has a 'has_key' method.  I assume that's why your object 
    # isn't being recognized as dictionary-like."  -- Mike Orr
    
    has_key = __contains__

flash = Flash()    

# ---------------------------------------------------------------------

def local_request(path, *args, **kwargs):
    """
    Map path to the proper object in the cherrypy.root tree and call it with the provided
    arguments. Generators are automatically iterated through and joined. This is automatically 
    included in views as $local_request.
    
    Return value is the result of the method call.
    """
    
    # ripped directly from CP _cphttptools.main(...) function.
    #   (BTW, it works here) -- deelan
    page_handler, object_path, virtual_path = _cphttptools.mapPathToObject(path)
    
    args = virtual_path + cherrypy.request.paramList
    r = page_handler(*args, **cherrypy.request.paramMap)
    #cherrypy.response.body = iterable(body)

    #r = _cphttptools.mapPathToObject(path)[0](*args, **kwargs)
    if isinstance(r, types.GeneratorType):
        return "".join(r)
    else:
        return r

def get_app_dir(dir=""):
    """
    Return the application root directory.
    Specify an extra directory to be joined with the root.
    """
    return os.path.join(os.path.dirname(os.path.dirname(os.path.normpath(cherrypy.root._module.__file__))), dir)
