# -*- coding: utf-8 -*-
"""
Subway built-in view filters.

Use them like this:

$escape($truncate(user.screen_name, length=3))

Note: several filters were lifted from the 
  Python's Django web framework. 

"""

from subway.util import markdown as _markdown
from subway.util import text

import cgi
import cherrypy
import urllib 

def _get_encoding(defalt_encoding='utf-8'):
    return cherrypy.config.get("subway.encoding", defalt_encoding)

# ---------------------------------------------------------------------

def escape(value):
    """
    Escape HTML <, > & and " symbols.
    """
    value = str(value) 
    return cgi.escape(value, quote=True)

def escape_url(value):
    """
    Escape given URL using percent-encoding.
    """    
    return urllib.quote(value)

# ---------------------------------------------------------------------

def truncate(value, length, extra=''):    
    """
    Truncate string up to length. 
    @length  number of characters to truncate after.
    @extra   add a string suffix when truncation happens
    """
    
    value = str(value).decode(_get_encoding())
        
    if len(value) > length:    
        extra = str(extra).decode(_get_encoding())                
        value = value[:length] + extra

    return value.encode(_get_encoding())


def truncate_words(value, length, extra=''):
    """
    Truncates a string after a certain number of words
    @length  number of words to truncate after.
    @extra   add a string suffix when truncation happens
    """
    
    value = str(value).decode(_get_encoding())
        
    words = value.split()
    if len(words) > length:            
        words = words[:length]
        
        if extra: 
            extra = str(extra).decode(_get_encoding())                
            words.append(extra)  
    
    return ' '.join(words).encode(_get_encoding())

# ---------------------------------------------------------------------

import re 

RE_NORMALIZE_NEWLINES = re.compile(r'\r\n|\r|\n') 
RE_BREAK_LINES = re.compile('\n{2,}')

def break_lines(value):
    """
    Converts newlines into <p> and <br />s
    """
    
    value = str(value)    
    
    value = RE_NORMALIZE_NEWLINES.sub('\n', value) 
    
    paras = RE_BREAK_LINES.split(value)
    paras = ['<p>%s</p>' % p.strip().replace('\n', '<br />') for p in paras]
    
    return '\n\n'.join(paras)
        

def strip_tags(value, keep_tags=()):
    """
    Strips HTML tags but keep_tags ones. By default 
    all tags are stripped.
    """

    p = text.HTMLSanitizer()
    
    # no acceptable elements and attributes
    p.acceptable_elements = keep_tags
    #@@ p.acceptable_attributes = safe_attrs
    
    p.feed(value)
    value = p.output()

    return value

def strip_unsafe_tags(value):
    """
    Strips unsafe HTML tags.
    """
    p = text.HTMLSanitizer()
        
    p.feed(value)
    value = p.output()

    return value


# ---------------------------------------------------------------------

import math
from datetime import datetime

CHUNKS = (
  (60 * 60 * 24 * 365, 'year'),
  (60 * 60 * 24 * 30, 'month'),
  (60 * 60 * 24, 'day'),
  (60 * 60, 'hour'),
  (60, 'minute')
)

def time_since(original, when):
    """
    Takes a datetime object, returns the time between then and now
    as a nicely formatted string, e.g "10 minutes"
    
    Adapted from http://blog.natbat.co.uk/archive/2003/Jun/14/time_since
    
    @original... 
    @ 
    """
            
    # get a timedelta object
    since = when - original

    for i, (seconds, name) in enumerate(CHUNKS):
        count = math.floor(since.seconds / seconds)
        if count != 0:
            break
    if count == 1:
        s = '1 %s' % name
    else:
        s = '%d %ss' % (count, name)
    if i + 1 < len(CHUNKS):
        # now get the second item
        seconds2, name2 = CHUNKS[i + 1]
        count2 = math.floor((since.seconds - (seconds * count)) / seconds2)
        if count2 != 0:
            if count2 == 1:
                s += ', 1 %s' % name2
            else:
                s += ', %d %ss' % (count2, name2)
    return s

#@@ time_until ???

def time_since_now(original):
    return time_since(original, datetime.now())

# ---------------------------------------------------------------------

# matches one or more consecutive whitespaces
RE_SLUG_WHITESPACE = re.compile("\s+")
# matches any non-alphanumeric character
RE_SLUG = re.compile("[^\w\s-]", re.IGNORECASE) 

def slugify(value):
    """
    Create an ASCII-only and URL friendly string.
    """
    
    # turn into unicode
    value = value.decode(_get_encoding()).lower()

    # take accented chars and turn into a closer 7bit ASCII equivalent
    value = text.translate(value)
    
    # turn subsequent whitespaces into -
    value = RE_SLUG_WHITESPACE.sub(u'-', value)
    # remove any non-alphanumeric character
    value = RE_SLUG.sub(u'', value)    

    # strip leading/trailing - (ex spaces)
    value = value.strip('-')

    # back to str and return 
    return value.lower().encode(_get_encoding())

# ---------------------------------------------------------------------

def format_size(value):
    """
    Format the value like a 'human-readable' 
    file size (i.e. 13 KB, 4.1 MB, 102 bytes, etc).
    """
                
    bytes = float(value)
    if bytes < 1024:
        return "%d byte%s" % (bytes, bytes != 1 and 's' or '')
    if bytes < 1024 * 1024:
        return "%.1f KB" % (bytes / 1024)
    if bytes < 1024 * 1024 * 1024:
        return "%.1f MB" % (bytes / (1024 * 1024))
    return "%.1f GB" % (bytes / (1024 * 1024 * 1024))

# ---------------------------------------------------------------------


def markdown(value):
    return _markdown.markdown(value)


# ---------------------------------------------------------------------

__all__ = ['escape', 'escape_url', 'break_lines', 'truncate', 'truncate_words', 
    'time_since', 'time_since_now', 'slugify', 'format_size', 'markdown', 'strip_tags', 
    'strip_unsafe_tags']

# ---------------------------------------------------------------------

def run_tests():
        
    assert escape('foo < > & "') == 'foo &lt; &gt; &amp; &quot;'

    # we but accents here and there to see if it blows
    assert truncate('A vèry long title', length=6) == 'A vèry'
    assert truncate('A vèry long title', length=6, extra='...') == 'A vèry...'
    
    assert break_lines('First para,\non the second line.\n\nAnother para.') == \
        '<p>First para,<br />on the second line.</p>\n\n<p>Another para.</p>'
    
    assert truncate_words('The qùick frog jumped over the lazy cow', length=3) == \
        'The qùick frog'

    assert strip_tags("The <em>qùick</em> frog umped over the <b>lazy</b> cow") == \
        'The qùick frog umped over the lazy cow'

    assert strip_tags("The <em>qùick</em> frog umped over the <b>lazy</b> cow", keep_tags=['b']) == \
        'The qùick frog umped over the <b>lazy</b> cow'

    assert strip_unsafe_tags("Evil!<script>window.location.href='http://evil.com'</script>") == \
        'Evil!'

    d = datetime(2005, 1, 1, 12, 00)
    now = datetime(2005, 1, 1, 13, 00)
    assert time_since(d, when=now) == '1 hour'

    now = datetime(2005, 1, 1, 12, 1)
    assert time_since(d, when=now) == '1 minute'

    now = datetime(2005, 1, 1, 12, 30)
    assert time_since(d, when=now) == '30 minutes'

    #@@ this fails
    #now = datetime(2005,2,1,12,00)
    #print  d|time_since(when=now)    
    #assert d|time_since(when=now) == '1 month, ...'

    assert slugify(' Àm i cool  ènough?  ') == 'am-i-cool-enough'
    assert slugify('999 .!?  #') == '999'
        
    assert format_size('988') == '988 bytes'
    assert format_size('1199') == '1.2 KB'
    assert format_size('1902222') == '1.8 MB'
    assert format_size('1785902345') == '1.7 GB'
    
    assert escape_url('http://example.com/~john connolly/') == 'http%3A//example.com/%7Ejohn%20connolly/'

if __name__ == '__main__':
    run_tests()
